if ($host.Version.Major -LT 5) {
	throw "PowerShell version outdated. Please update. Required version: 5.1.x.x."
}

if ($host.Version.Minor -LT 1) {
	if ($host.Version.Major -LE 5) {
		throw "PowerShell version outdated. Please update. Required version: 5.1.x.x."
	}
}

Add-Type @"
using System;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
public class ServerCertificateValidationCallback {
    public static void DoNotIgnore(Object[] certificates) {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
                foreach(Object trustedCertificate in certificates) {
                    if (trustedCertificate.Equals(certificate)) {
                        return true;
                    }
                }
				return false;
			};
	}

	public static void Ignore() {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
				return true;
			};
	}

	public static void AcceptOnly(X509Certificate cert, Object[] certificates) {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
				if (cert.Equals(certificate)) {
                    return true;
                }

                foreach(Object trustedCertificate in certificates) {
                    if (trustedCertificate.Equals(certificate)) {
                        return true;
                    }
                }

                return false;
			};
	}
}
"@

$script:Version='2.7.2'
$script:APIVersion=0
$script:Address=''
$script:Port=443
$script:Credential=$null
$script:CertificateCheckDisabled=$false

# Disable Certificate check (at your own risk!)
# This will not only disable the certificate check for anynode requests, but for
# all requests.
function Disable-CertificateCheck {
    $script:CertificateCheckDisabled=$true
	[ServerCertificateValidationCallback]::Ignore()
}

function Enable-CertificateCheck {
    $script:CertificateCheckDisabled=$false
	[ServerCertificateValidationCallback]::DoNotIgnore((Get-ChildItem 'Cert:/CurrentUser/Root/'))
}

function Initialize-anynodeAddress {
    [CmdletBinding()]
    Param(
        [Parameter(mandatory=$true)]
        [string] $Address
    )

    if ($Address -eq $null -OR $Address -eq '') {
        echo 'Please enter an address.'
        return
    }

    $script:Address=$Address
    echo "Address to communicate with anynode is set to: $Address"
}

function Reset-anynodeAddress {
    $script:Address=''
    echo 'Reset address used to communicate with anynode.'
}

function Initialize-anynodePort {
    [CmdletBinding()]
    Param(
        [Parameter(mandatory=$true)]
        [string] $Port
    )

    if ($Port -eq $null -OR $Port -le 0) {
        echo 'Please enter a valid port.'
        return
    }

    $script:Port=$Port
    echo "Port to communicate with anynode is set to: $Port"
}

function Reset-anynodePort {
    $script:Port=443
    echo 'Reset port used to communicate with anynode to value: 443'
}

function Initialize-anynodeCredential {
    [CmdletBinding(DefaultParameterSetName='cred')]
    Param(
        [Parameter(mandatory=$true,ParameterSetName='cred')]
        [PSCredential] $Credential,
        [Parameter(mandatory=$true,ParameterSetName='userpass')]
        [string] $Username,
        [Parameter(mandatory=$true,ParameterSetName='userpass')]
        [string] $Password
    )

    if ($Credential -ne $null) {
        $script:Credential=$Credential
    } elseif ($Username -ne $null) {
        try {
            $SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
	        $script:Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
        } catch {
            echo 'Could not set default credential.'
            return
        }
    }

    echo 'Set default credential.'
}

function Reset-anynodeCredential {
    $script:Credential=$null
    echo 'Reset default credential.'
}

# During this session, you can only communicate with the given interface.
function Get-anynodeWebserverCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $CertificateFilePath
	)

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

    Disable-CertificateCheck

	try {
		$Request=[System.Net.WebRequest]::Create('https://'+$Address+':'+$Port+'/')
		$Request.GetResponse().Dispose()
		$Certificate=$Request.ServicePoint.Certificate

		if (!($CertificateFilePath -eq $null) -AND !($CertificateFilePath -eq '')) {
			$CertBytes=$Certificate.Export([System.Security.Cryptography.X509Certificates.X509ContentType]::Cert)
			Set-Content -Value $CertBytes -Encoding Byte -Path $CertificateFilePath
		}

		[ServerCertificateValidationCallback]::AcceptOnly($Certificate, (Get-ChildItem 'Cert:/CurrentUser/Root/'))
	} catch {
        Enable-CertificateCheck
		$_.Exception
		return
	}
}

function Import-anynodeWebserverCertificate {
	[CmdletBinding()]
	Param(
		[Parameter(mandatory=$true)]
		[string] $CertificatePath
	)

	$Certificate
	try {
		$Certificate=[X509Certificate]::CreateFromCertFile($CertificatePath)
	} catch {
		$_.Exception
		return
	}

	[ServerCertificateValidationCallback]::AcceptOnly($Certificate, (Get-ChildItem 'Cert:/CurrentUser/Root/'))
}

# Invokes
#	Download
function Invoke-anynodeDownload {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	try {
		$Request=Invoke-WebRequest -Method GET -URI $URI -Headers $Headers

		$Request
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                $Request=Invoke-WebRequest -Method GET -URI $URI -Headers $Headers

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
        }
	}
}

#	Request
function Invoke-anynodeRequest {
	[CmdletBinding()]
	Param(
		[Parameter(mandatory=$true)]
		[ValidateSet('Get','Put','Post','Delete')]
		[string] $Method,
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter()]
		[string] $Body,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	try {
		if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {
			$Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers -Body $Body

		} else {
			$Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers
		}

		$Request
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {
			        $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers -Body $Body

		        } else {
			        $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers
		        }

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
        }
	}
}

#	Upload
function Invoke-anynodeUpload {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter(mandatory=$true)]
		[Object[]] $Filenames,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$Boundary=[System.Guid]::NewGuid().ToString()
	$ContentType='multipart/form-data; boundary="'+$Boundary+'"'

	$NL="`r`n";

	$Body=[byte[]]::new(0)

	foreach ($Filename in $Filenames) {
		$Bytes=[System.IO.File]::ReadAllBytes($Filename)

		$Head=(
			"--$Boundary",
		    "Content-Disposition: form-data; name=`"file`"; filename=`"$Filename`"",
		    "Content-Type: application/octet-stream$NL$NL"
		) -join $NL

		$Body+=[Text.Encoding]::UTF8.GetBytes($Head)
		$Body+=$Bytes

		$Foot="$NL--$Boundary--$NL"
		$Body+=[Text.Encoding]::UTF8.GetBytes($Foot)
	}

	$tmp=New-TemporaryFile
	$tmpName=$tmp.FullName
	[IO.File]::WriteAllBytes($tmpName,$Body)

	try {
		Invoke-RestMethod -Method PUT -Uri $URI -Headers $Headers -ContentType $ContentType -InFile $tmpName
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                Invoke-RestMethod -Method PUT -Uri $URI -Headers $Headers -ContentType $ContentType -InFile $tmpName

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
        }
	} finally {
		Remove-Item $tmpName -Force	
	}
}

# Active Sessions
#	Get
function Get-anynodeActiveSessions {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /activeSessions/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Backends
#	Get
function Get-anynodeBackends {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /backends/get -Credential $Credential -APIVersion $APIVersion -Username $Username -Password $Password
}

# Config
#	Get
function Get-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Filename -eq '')) {
		$Parameters.Add("filename",$Filename)
	}

	$Request=Invoke-anynodeDownload -Address $Address -Port $Port -URITail /anynodeConfig/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	if ($Filename -eq '') {
		if ($Request.Headers -eq $null) {
			echo $Request
			return
		}

		$ContentDisposition=$Request.Headers."Content-Disposition"
		if ($ContentDisposition -eq $null) {
			echo 'Could not find Content-Disposition header.'
			return
		}

		$split=$ContentDisposition -split 'filename="'
		if (($split[1] -eq $null) -Or ($split[1] -eq '')) {
			echo 'Could not find filename.'
			return
		}

		$Filename=($split[1] -split '"')[0]
	}

	try {
		[Text.Encoding]::Default.GetString($Request.Content) | Out-File $Filename -Encoding default
		echo 'Saved file at "'$Filename'".'
	} catch {
		echo $_.Exception
	}
}

#	Import
function Import-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[switch] $CommitProtection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=''
	try {
		$Body=[IO.File]::ReadAllText($Filepath)
	} catch {
		echo 'Could not find file.'
		return
	}

	$Parameters=@{}
	if ($CommitProtection) {
		$Parameters.Add("commitProtection",$true)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /anynodeConfig/set -Filenames @($Filepath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Push
function Push-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /anynodeConfig/commit -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Reset
function Reset-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /anynodeConfig/revert -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Dashboard
#	Get
function Get-anynodeDashboard {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('sipNodes','sfbUcmaNodes','ldapConnections')]
		[psobject] $List,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($List -eq $null)) {
		$Parameters.Add("list",$List)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /dashboard/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# License
#	Get
function Get-anynodeLicenses {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /licenses/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeLicenses {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /licenses/clear -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeLicense {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=''
	try {
		$Body=[IO.File]::ReadAllText($Filepath)
	} catch {
		echo 'Could not find file.'
		return
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /licenses/importFile -Filenames @($Filepath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# LOD
#	Import
function Import-anynodeLOD {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $LODKey,
		[Parameter()]
		[string] $LicensingDevice,
		[Parameter(mandatory=$true)]
		[string] $EmailAddress,
		[Parameter(mandatory=$true)]
		[string] $CompanyName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add("lodKey",$LODKey)

	if (!($LicensingDevice -eq '')) {
		$Body.Add("licensingDevice",$LicensingDevice)
	}

	$Body.Add("emailAddress",$EmailAddress)

	$Body.Add("companyName",$CompanyName)

	$license=Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /licenses/importLod -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	$Identifier=$license.identifier
	
	if ($Identifier -eq $null) {
		$license
		return
	}

	$LicensesLoaded=Get-anynodeLicenses -Address $Address -Port $Port -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	$Result

	foreach ($LicenseLoaded in $LicensesLoaded) {
		if ($LicenseLoaded.identifier -eq $Identifier) {
			$Result=$LicenseLoaded
		}
	}

	if ($Result -eq $null) {
		$license
	} else {
		$Result
	}
}

# Maintenance Mode
#	Get
function Get-anynodeMaintenanceMode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /maintenanceMode/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeMaintenanceMode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ModeActive,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	
	if (!($ModeActive -eq $null)) {
		$Body.Add("modeActive",$ModeActive)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /maintenanceMode/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Module Version
#	Get
function Get-anynodeModuleVersion {
	$script:Version
}

# Network Controller
#	Get
function Get-anynodeNetworkControllers {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /networkControllers/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Interfaces
function Get-anynodeNetworkInterfaces {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /networkInterfaces/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node
#	Get
function Get-anynodeNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $IncludeNetworkInformation,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}

    if ($IncludeNetworkInformation -ne $null) {
        $Parameters.Add("includeNetworkInformation",$IncludeNetworkInformation)
    }

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Provider Profiles
function Get-anynodeProviderNodeProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/providerProfiles/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Provider
function Add-anynodeProviderNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateNatTraversal,
		[Parameter()]
		[string] $ExternalHost,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NatDeviceHasFixedPortMapping,
		[Parameter()]
		[int] $FixedUdpTcpPort,
		[Parameter()]
		[int] $FixedTlsPort,
		[Parameter()]
		[int] $FixedUdpPortRangeStart,
		[Parameter()]
		[int] $FixedTcpPortRangeStart,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NodeViaSipTrunking,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[string] $ProxyUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $DefineAssertedUri,
		[Parameter()]
		[string] $AssertedUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeProxyInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($ActivateNatTraversal -eq $null)) {
		$Body.Add("activateNatTraversal",$ActivateNatTraversal)
	}
	if ($ActivateNatTraversal) {
		$Body.Add("externalHost",$ExternalHost)

		if (!($NatDeviceHasFixedPortMapping -eq $null)) {
			$Body.Add("natDeviceHasFixedPortMapping",$NatDeviceHasFixedPortMapping)
		}
		if ($NatDeviceHasFixedPortMapping) {
			$Body.Add("fixedUdpTcpPort",$FixedUdpTcpPort)
			$Body.Add("fixedTlsPort",$FixedTlsPort)
			$Body.Add("fixedUdpPortRangeStart",$FixedUdpPortRangeStart)
			$Body.Add("fixedTcpPortRangeStart",$FixedTcpPortRangeStart)
		}
	}

	if (!($NodeViaSipTrunking -eq $null)) {
		$Body.Add("nodeViaSipTrunking",$NodeViaSipTrunking)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}

	if (!($ClientCredentials -eq $null)) {
		$Body.Add("credentials",$ClientCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($ProxyUri -eq $null) -AND !($ProxyUri -eq '')) {
		$Body.Add("proxyUri",$ProxyUri)
	}

	if (!($DefineAssertedUri -eq $null)) {
		$Body.Add("defineAssertedUri",$DefineAssertedUri)
	}
	if ($DefineAssertedUri) {
		$Body.Add("assertedUri",$AssertedUri)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addProvider -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get System Profiles
function Get-anynodeSystemNodeProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/systemProfiles/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add System
function Add-anynodeSystemNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet("REGISTRAR","REGISTRATION","TRUNK")]
		[psobject] $SipInterconnection,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateServerAuthentication,
		[Parameter()]
		[psobject] $ServerCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[psobject] $SipRegistrar,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeOwnSubnetInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($SipInterconnection -eq $null)) {
		$Body.Add("sipInterconnection",$SipInterconnection)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}
	if (!($ClientCredentials -eq $null)) {
		$Body.Add("clientCredentials",$ClientCredentials)
	}

	if (!($ActivateServerAuthentication -eq $null)) {
		$Body.Add("activateServerAuthentication",$ActivateServerAuthentication)
	}
	if (!($ServerCredentials -eq $null)) {
		$Body.Add("serverCredentials",$ServerCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($SipRegistrar -eq $null)) {
		$Body.Add("sipRegistrar",$SipRegistrar)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add MS Teams Tenant
function Add-anynodeMsTeamsTenantNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewTeamsCarrierNode=$false,
		[Parameter()]
		[string] $TeamsCarrierNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[string] $SbcFqdn,
		[Parameter()]
		[string] $TenantFqdn,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Profile=Get-anynodeSystemNodeProfiles -Address $Address -Port $Port -Credential $Credential -Username $Username -Password $Password | Where-Object -Property "displayName" -EQ "Microsoft Teams Direct Routing Carrier"

	$Parameters=@{profile=$Profile.id}

	$Body=@{}

	$Body.Add("createNewTeamsCarrierNode",$CreateNewTeamsCarrierNode)
	if ($CreateNewTeamsCarrierNode) {

		$Body.Add("createNewNetworkController",$CreateNewNetworkController)
		if ($CreateNewNetworkController) {
			if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
				$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
			}
			$Body.Add("networkControllerInterface",$NetworkControllerInterface)
			$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
			$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
			if (!($EnableReverseDnsLookup -eq $null)) {
				$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
			}
		}

		if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
			$Body.Add("networkControllerId",$NetworkControllerId)
		}
	
		if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
			$Body.Add("udpTcpPort",$UdpTcpPort)
		}
	
		if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
			$Body.Add("tlsPort",$TlsPort)
		}
	
		if (!($UdpPortRange -eq $null)) {
			$Body.Add("udpPortRange",$UdpPortRange)
		}
	
		if (!($TcpPortRange -eq $null)) {
			$Body.Add("tcpPortRange",$TcpPortRange)
		}

		if (!($SbcFqdn -eq $null) -AND !($SbcFqdn -eq '')) {
			$Body.Add("sbcFqdn",$SbcFqdn)
		}
	}

	if (!($TeamsCarrierNodeId -eq $null) -AND !($TeamsCarrierNodeId -eq '')) {
		$Body.Add("teamsCarrierNodeId",$TeamsCarrierNodeId)
	}

	if (!($TenantFqdn -eq $null) -AND !($TenantFqdn -eq '')) {
		$Body.Add("tenantFqdn",$TenantFqdn)
	}

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Incoming Manipulations
function Get-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/incomingManipulations/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Incoming Manipulations
function Set-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingManipulations/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Outgoing Manipulations
function Get-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Outgoing Manipulations
function Set-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $Force,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($Force) {
		$Parameters.Add("force",$true)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get MS Teams Connectivities
function Get-anynodeMsTeamsCarriers {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/msTeamsCarriers/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Certificates
#	Get
function Get-anynodeNodeCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/certificates/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[string] $CertificateFilePath,
		[Parameter()]
		[string] $KeyFilePath,
		[Parameter()]
		[string] $CertificatePassword,
		[Parameter()]
		[ValidateSet("CERTIFICATES_AND_PRIVATE_KEY","CERTIFICATES_ONLY","PRIVATE_KEY_ONLY")]
		[string] $Mode='CERTIFICATES_AND_PRIVATE_KEY',
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if (!($CertificatePassword -eq '')) {
		$Parameters.Add("password",$CertificatePassword)
	}

	if (!($CertificateFilePath -eq '')) {
		$RequestParams=$Parameters.Clone()
		if ($Mode -eq 'CERTIFICATES_AND_PRIVATE_KEY') {
			$RequestParams.Add("mode",'CERTIFICATES_ONLY')
		} elseif (!($Mode -eq '')) {
			$RequestParams.Add("mode",$Mode)
		}

		Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $RequestParams -Filenames @($CertificateFilePath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	}

	if (!($KeyFilePath -eq '')) {
		$RequestParams=$Parameters.Clone()
		if ($Mode -eq 'CERTIFICATES_AND_PRIVATE_KEY') {
			$RequestParams.Add("mode",'PRIVATE_KEY_ONLY')
		} elseif (!($Mode -eq '')) {
			$RequestParams.Add("mode",$Mode)
		}

		Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $RequestParams -Filenames @($KeyFilePath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	}

# This doesn't work because anynode only accepts one file at a time.
#	$Files=@{}
#	if (!($CertificateFilePath -eq '')) {
#		$Files.Add(1,$CertificateFilePath)
#	}
#
#	if (!($KeyFilePath -eq '')) {
#		$Files.Add(2,$KeyFilePath)
#	}
#
#	if ($Files.Count -eq 0) {
#		echo 'Path(s) to Certificate and/or Private Key must be given for this call to work.'
#		return
#	}
#
#	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $Parameters -Filenames $Files.Values -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Trusted Certificates
#	Get
function Get-anynodeNodeTrustedCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/trustedCertificates/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeTrustedCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $CertificateFilePath,
		[Parameter()]
		[string] $CertificatePassword,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if (!($CertificatePassword -eq '')) {
		$Parameters.Add("password",$CertificatePassword)
	}

	$Files=@{}
	if (!($CertificateFilePath -eq '')) {
		$Files.Add(1,$CertificateFilePath)
	}

	if ($Files.Count -eq 0) {
		echo 'Path to a Certificate must be given for this call to work.'
		return
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/trustedCertificates/add -Parameters $Parameters -Filenames $Files.Values -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeTrustedCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/trustedCertificates/clear -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node FQDN
#	Get
function Get-anynodeNodeFQDN {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/tenants/fqdn/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeFQDN {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $NewFQDN,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/tenants/fqdn/set -Body $NewFQDN -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node override hostname
#	Get
function Get-anynodeNodeOverrideHostname {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/overrideHostname/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeOverrideHostname {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $NewOverrideHostname,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/overrideHostname/set -Body $NewOverrideHostname -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Port
#	Get
function Get-anynodeNodePort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/port/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodePort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[int] $NewPort,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/port/set -Body $NewPort -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node TLS Port
#	Get
function Get-anynodeNodeTlsPort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/tlsPort/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeTlsPort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[int] $NewPort,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/tlsPort/set -Body $NewPort -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Transport Connection
#	Get
function Get-anynodeNodeTransportConnections {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/transportConnections/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject]$MatchingCondition,
		[Parameter(mandatory=$true)]
		[psobject]$TransportConnectionId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("transportConnectionId",$TransportConnectionId)
	$Body.Add("matchingCondition",$MatchingCondition)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/transportConnections/add -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeTransportConnections {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/transportConnections/clear -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Transport Connections
#	Get
function Get-anynodeTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/get -Parameters @{id=$Id} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Authentication
function Get-anynodeTransportConnectionAuthentication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/authentication/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Authentication
function Set-anynodeTransportConnectionAuthentication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $TransportConnectionUsername,
		[Parameter()]
		[string] $TransportConnectionPassword,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($TransportConnectionUsername -eq '')) {
		$Body.Add('username',$TransportConnectionUsername)
	}

	if (!($TransportConnectionPassword -eq '')) {
		$Body.Add('password',$TransportConnectionPassword)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/authentication/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Asserted URI
function Get-anynodeTransportConnectionAssertedURI {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/assertedUri/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Asserted URI
function Set-anynodeTransportConnectionAssertedURI {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $URI,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($URI -eq '')) {
		$Body.Add('iri',$URI)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add('displayName',$DisplayName)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/assertedUri/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Address of Record
function Get-anynodeTransportConnectionAddressOfRecord {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/addressOfRecord/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Address of Record
function Set-anynodeTransportConnectionAddressOfRecord {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $URI,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($URI -eq '')) {
		$Body.Add('iri',$URI)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add('displayName',$DisplayName)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/addressOfRecord/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
#		Standard
function Add-anynodeStandardTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[string] $SipNodeId,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $UseAProxyServer,
		[Parameter()]
		[psobject] $Proxy,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $RegistrationEnabled,
		[Parameter()]
		[string] $SipDomain,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $AuthenticationEnabled,
		[Parameter()]
		[psobject] $AuthenticationPolicy,
		[Parameter()]
		[string] $AssertedURI,
		[Parameter()]
		[string] $AssertedDisplayName,
		[Parameter()]
		[string] $ConditionId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($SipNodeId -eq '')) {
		$Body.Add('sipNodeId',$SipNodeId)
	}

	if (!($Proxy -eq $null)) {
		$Body.Add('proxy',$Proxy)
	}

	if ($RegistrationEnabled) {
		$Registration=@{}
		if (!($SipDomain -eq '')) {
			$Registration.Add('sipDomain',$SipDomain)
		}

		if (!($AddressOfRecord -eq '')) {
			$Registration.Add('addressOfRecord',$AddressOfRecord)
		}

		$Body.Add('registration',$Registration)
	}

	if (!($AuthenticationPolicy -eq '')) {
		$Body.Add('credentials',$AuthenticationPolicy)
	}

	if (!($AssertedURI -eq '')) {
		$AssertedAddress=@{iri=$AssertedURI}

		if (!($AssertedDisplayName -eq '')) {
			$AssertedAddress.Add('displayName',$AssertedDisplayName)
		}

		$Body.Add('assertedAddress',$AssertedAddress)
	}

	if (!($ConditionId -eq '')) {
		$Body.Add('conditionId',$ConditionId)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/addStandard -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{displayName=$DisplayName} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Object
#	Get
#		Single
function Get-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Multiple
function Get-anynodeObjects {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Sort,
		[Parameter()]
		[string] $DisplayName,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($Sort) {
		$Parameters.Add("sort",$Sort)
	}

	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	if (!($DisplayName -eq '')) {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/sorts -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter(mandatory=$true)]
		[psobject] $Object,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=($Object | ConvertTo-Json -Depth 99)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /objects/set -Body $Body -Parameters @{id=$Id} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $RemoveSubsequent,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}
	if (!($RemoveSubsequent -EQ $null)) {
		$Parameters.Add("removeSubsequent",$RemoveSubsequent)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /objects/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[ValidateSet("SKIP","OVERRIDE","CLONE","CANCEL")]
		[psobject] $DuplicateOption,
		[Parameter()]
		[string] $CloneComment,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($DuplicateOption -eq $null)) {
		$Parameters.Add("action",$DuplicateOption)
	}
	if (!($CloneComment -eq '')) {
		$Parameters.Add("cloneComment",$CloneComment)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /objects/import -Filenames @($Filepath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Export
function Export-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludePrivateData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}
	if (!($IncludePrivateData -eq $null)) {
		$Parameters.Add("includePrivateData",$IncludePrivateData)
	}

	$Request=Invoke-anynodeDownload -Address $Address -Port $Port -URITail /objects/export -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	if ($Request.Content -eq $null) {
		$Request
		return
	}

	if ($Filename -eq '') {
		if ($Request.Headers -eq $null) {
			echo $Request
			return
		}

		$ContentDisposition=$Request.Headers."Content-Disposition"
		if ($ContentDisposition -eq $null) {
			echo 'Could not find Content-Disposition header.'
			return
		}

		$split=$ContentDisposition -split 'filename="'
		if (($split[1] -eq $null) -Or ($split[1] -eq '')) {
			echo 'Could not find filename.'
			return
		}

		$Filename=($split[1] -split '"')[0]
	}

	[Text.Encoding]::Default.GetString($Request.Content) | Out-File $Filename -Encoding default

	echo 'Saved file at "'$Filename'".'
}

# Routing Domain
#	Get
function Get-anynodeRoutingDomains {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeSimpleRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter(mandatory=$true)]
		[string] $Node1,
		[Parameter(mandatory=$true)]
		[string] $Node2,
		[Parameter()]
		[string] $PrefixTo1,
		[Parameter()]
		[string] $PrefixTo2,
		[Parameter()]
		[string] $RoutingForwardProfile,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=0,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{"node1"=$Node1}
	$Body.Add("node2",$Node2)

	if (!($PrefixTo1 -eq '')) {
		$Body.Add("prefixTo1",$PrefixTo1)
	}

	if (!($PrefixTo2 -eq '')) {
		$Body.Add("prefixTo2",$PrefixTo2)
	}

	if (!($RoutingForwardProfile -eq '')) {
		$Body.Add("routingForwardProfile",$RoutingForwardProfile)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/add -Body ($Body | ConvertTo-Json) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Add-anynodeRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $SourceNodes,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("displayName",$DisplayName)
    $Body.Add("sourceNodes",$SourceNodes)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Routes
#	Add
function Add-anynodeEstablishmentRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseRouteSupervision,
        [Parameter()]
        [string] $SupervisionId,
        [Parameter()]
        [string] $DisplayName,
        [Parameter()]
        [Object[]] $Filters,
        [Parameter()]
        [ValidateSet('ROUTE','IGNORE','REJECT','REDIRECT','ESTABLISH_PARALLEL')]
        [string] $Type,
        [Parameter()]
        [string] $DestinationNodeId,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $SkipIfNodeNotOperational,
        [Parameter()]
        [ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
        [string] $RoutingForwardProfileSelection,
        [Parameter()]
        [string] $RoutingForwardProfileDisplayName,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $AdjustRoutingForwardProfileSettings,
        [Parameter()]
        [ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
        [string] $RoutingForwardProfileMode,
        [Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
        [ValidateSet('SUCCESS','DIAL_STRING','NETWORK_PERMISSION','NETWORK_CONGESTION','NETWORK_EQUIPMENT','USER_BUSY','USER_REDIRECTED','USER_NOT_RESPONDING','USER_NOT_SELECTED','USER_REJECTED','USER_TERMINATED','MEDIA_NEGOTIATION','ERROR','DOMAIN_SPECIFIC_0','DOMAIN_SPECIFIC_1','DOMAIN_SPECIFIC_2','DOMAIN_SPECIFIC_3','DOMAIN_SPECIFIC_4','DOMAIN_SPECIFIC_5','DOMAIN_SPECIFIC_6','DOMAIN_SPECIFIC_7','DOMAIN_SPECIFIC_8','DOMAIN_SPECIFIC_9','DOMAIN_SPECIFIC_10','DOMAIN_SPECIFIC_11','DOMAIN_SPECIFIC_12','DOMAIN_SPECIFIC_13','DOMAIN_SPECIFIC_14','DOMAIN_SPECIFIC_15','DOMAIN_SPECIFIC_16','DOMAIN_SPECIFIC_17','DOMAIN_SPECIFIC_18','DOMAIN_SPECIFIC_19')]
        [string] $RejectStatus,
        [Parameter()]
        [string] $ParallelCallDialStringDirectoryId,
        [Parameter()]
        [PSCustomObject] $SourceDialStringRewriting,
        [Parameter()]
        [PSCustomObject] $DestinationDialStringRewriting,
        [Parameter()]
        [Object[]] $ParallelCalls,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","establishment")

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }

    echo $Filters

    $Establishment=@{}

    if (!($Type -eq $null)) {
        $Establishment.Add("type",$Type)
    }

    if (!($DestinationNodeId -eq $null)) {
        $Establishment.Add("destinationNodeId",$DestinationNodeId)
    }

    if (!($SkipIfNodeNotOperational -eq $null)) {
        $Establishment.Add("skipIfNodeNotOperational",$SkipIfNodeNotOperational)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $Establishment.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $Establishment.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $Establishment.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $Establishment.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $Establishment.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    if (!($RejectStatus -eq $null)) {
        $Establishment.Add("rejectStatus",$RejectStatus)
    }

    if (!($ParallelCallDialStringDirectoryId -eq $null)) {
        $Establishment.Add("parallelCallDialStringDirectoryId",$ParallelCallDialStringDirectoryId)
    }

    if (!($SourceDialStringRewriting -eq $null)) {
        $Establishment.Add("sourceDialStringRewriting",$SourceDialStringRewriting)
    }

    if (!($DestinationDialStringRewriting -eq $null)) {
        $Establishment.Add("destinationDialStringRewriting",$DestinationDialStringRewriting)
    }

    if (!($ParallelCalls -eq $null)) {
        $Establishment.Add("parallelCalls",$ParallelCalls)
    }

    $Body.Add("establishment",$Establishment)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Add-anynodePathReplacementRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseRouteSupervision,
        [Parameter()]
        [string] $SupervisionId,
        [Parameter()]
        [string] $DisplayName,
        [Parameter()]
        [Object[]] $Filters,
        [Parameter()]
        [ValidateSet('ACCEPT','REJECT')]
        [string] $Type,
        [Parameter()]
        [string] $DestinationNodeId,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $SkipIfNodeNotOperational,
        [Parameter()]
        [ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
        [string] $RoutingForwardProfileSelection,
        [Parameter()]
        [string] $RoutingForwardProfileDisplayName,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $AdjustRoutingForwardProfileSettings,
        [Parameter()]
        [ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
        [string] $RoutingForwardProfileMode,
        [Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","pathReplacement")

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }

    $PathReplacement=@{}

    if (!($Type -eq $null)) {
        $PathReplacement.Add("type",$Type)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $PathReplacement.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $PathReplacement.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $PathReplacement.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $PathReplacement.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $PathReplacement.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    $Body.Add("pathReplacement",$PathReplacement)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Routing Forward Profile
#	Get
function Get-anynodeRoutingForwardProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingForwardProfiles/get -Parameters @{completeData=$CompleteData} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeRoutingForwardProfile {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RemoveAnynodeFromSignalingPath,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $EarlyMediaPassThrough,
		[Parameter()]
		[string] $TelephonyForwardingStackId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RecordIncoming,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RecordOutgoing,
		[Parameter()]
		[string] $MediaRecorderId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AnnouncementIncoming,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AnnouncementOutgoing,
		[Parameter()]
		[string] $AnnouncementMediaSourceId,
		[Parameter()]
		[string] $MediaNegotiationForwarderId,
		[Parameter()]
		[PSCustomObject] $SignalingForwardingFromCallingToCalled,
		[Parameter()]
		[PSCustomObject] $SignalingForwardingFromCalledToCalling,
		[Parameter()]
		[string] $MediaTranscodingOptionsId,
		[Parameter()]
		[ValidateSet('TRANSCODING','BYPASS','PASSTHROUGH','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
		[string] $MediaMode,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add('displayName',$DisplayName)

	if (!($RemoveAnynodeFromSignalingPath -eq $null)) {
		$Body.Add('removeAnynodeFromSignalingPath',$RemoveAnynodeFromSignalingPath)
	}

	if (!($EarlyMediaPassThrough -eq $null)) {
		$Body.Add('earlyMediaPassThrough',$EarlyMediaPassThrough)
	}

	if (!($TelephonyForwardingStackId -eq '')) {
		$Body.Add('telephonyForwardingStackId',$TelephonyForwardingStackId)
	}

	if (!($RecordIncoming -eq $null)) {
		$Body.Add('recordIncoming',$RecordIncoming)
	}

	if (!($RecordOutgoing -eq $null)) {
		$Body.Add('recordOutgoing',$RecordOutgoing)
	}

	if (!($MediaRecorderId -eq '')) {
		$Body.Add('mediaRecorderId',$MediaRecorderId)
	}

	if (!($AnnouncementIncoming -eq $null)) {
		$Body.Add('announcementIncoming',$AnnouncementIncoming)
	}

	if (!($AnnouncementOutgoing -eq $null)) {
		$Body.Add('announcementOutgoing',$AnnouncementOutgoing)
	}

	if (!($AnnouncementMediaSourceId -eq '')) {
		$Body.Add('announcementMediaSourceId',$AnnouncementMediaSourceId)
	}

	if (!($MediaNegotiationForwarderId -eq '')) {
		$Body.Add('mediaNegotiationForwarderId',$MediaNegotiationForwarderId)
	}

	if (!($SignalingForwardingFromCallingToCalled -eq $null)) {
		$Body.Add('signalingForwardingFromCallingToCalled',$SignalingForwardingFromCallingToCalled)
	}

	if (!($SignalingForwardingFromCalledToCalling -eq $null)) {
		$Body.Add('signalingForwardingFromCalledToCalling',$SignalingForwardingFromCalledToCalling)
	}

	if (!($MediaTranscodingOptionsId -eq '')) {
		$Body.Add('mediaTranscodingOptionsId',$MediaTranscodingOptionsId)
	}

	if (!($MediaMode -eq '')) {
		$Body.Add('mediaMode',$MediaMode)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingForwardProfiles/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Sorts
#	Get
function Get-anynodeSorts {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/sorts -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# SystemIds
#	Get
function Get-anynodeSystemIds {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /licenses/systemIds/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Trace
#	Get
function Get-anynodeTrace {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Filename -eq '')) {
		$Parameters.Add("filename",$Filename)
	}

	$Request=Invoke-anynodeDownload -Address $Address -Port $Port -URITail /trace/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	if ($Request.Content -eq $null) {
		$Request
		return
	}

	if ($Filename -eq '') {
		if ($Request.Headers -eq $null) {
			echo $Request
			return
		}

		$ContentDisposition=$Request.Headers."Content-Disposition"
		if ($ContentDisposition -eq $null) {
			echo 'Could not find Content-Disposition header.'
			return
		}

		$split=$ContentDisposition -split 'filename="'
		if (($split[1] -eq $null) -Or ($split[1] -eq '')) {
			echo 'Could not find filename.'
			return
		}

		$Filename=($split[1] -split '"')[0]
	}

	[Text.Encoding]::Default.GetString($Request.Content) | Out-File $Filename -Encoding default

	echo 'Saved file at "'$Filename'".'
}

# Tracing
#	Get
function Get-anynodeTracing {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /tracing/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeTracing {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('FILE','SPLIT')]
		[string] $Selection,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $TraceActive,
		[Parameter()]
		[pscustomobject] $Options,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludePrivateData,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	if (!($Selection -eq '')) {
		$Body.Add("selection",$Selection)
	}

	if (!($TraceActive -eq $null)) {
		$Body.Add("traceActive",$TraceActive)
	}

	if (!($Options -eq $null)) {
		$Body.Add("options",($Options | ConvertTo-Json -Depth 99 | ConvertFrom-Json))
	}

	if (!($IncludePrivateData -eq $null)) {
		$Body.Add("includePrivateData",$IncludePrivateData)
	}

	if (!($IncludeDefaults -eq $null)) {
		$Body.Add("includeDefaults",$IncludeDefaults)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /tracing/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Version
#	Get
function Get-anynodeVersion {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('anynode','java','ucma','administration','monitor','frontend')]
		[string] $Product,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Product -eq '')) {
		$Parameters.Add("product",$Product)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /version/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}
# SIG # Begin signature block
# MIIb6AYJKoZIhvcNAQcCoIIb2TCCG9UCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQU+XXHB8c2tWPi3yFFVjU3Thjq
# 0TOgghaNMIIDXzCCAkegAwIBAgILBAAAAAABIVhTCKIwDQYJKoZIhvcNAQELBQAw
# TDEgMB4GA1UECxMXR2xvYmFsU2lnbiBSb290IENBIC0gUjMxEzARBgNVBAoTCkds
# b2JhbFNpZ24xEzARBgNVBAMTCkdsb2JhbFNpZ24wHhcNMDkwMzE4MTAwMDAwWhcN
# MjkwMzE4MTAwMDAwWjBMMSAwHgYDVQQLExdHbG9iYWxTaWduIFJvb3QgQ0EgLSBS
# MzETMBEGA1UEChMKR2xvYmFsU2lnbjETMBEGA1UEAxMKR2xvYmFsU2lnbjCCASIw
# DQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAMwldpB5BngiFvXAg7aEyiie/QV2
# EcWtiHL8RgJDx7KKnQRfJMsuS+FggkbhUqsMgUdwbN1k0ev1LKMPgj0MK66X17YU
# hhB5uzsTgHeMCOFJ0mpiLx9e+pZo34knlTifBtc+ycsmWQ1z3rDI6SYOgxXG71uL
# 0gRgykmmKPZpO/bLyCiR5Z2KYVc3rHQU3HTgOu5yLy6c+9C7v/U9AOEGM+iCK65T
# pjoWc4zdQQ4gOsC0p6Hpsk+QLjJg6VfLuQSSaGjlOCZgdbKfd/+RFO+uIEn8rUAV
# SNECMWEZXriX7613t2Saer9fwRPvm2L7DWzgVGkWqQPabumDk3F2xmmFghcCAwEA
# AaNCMEAwDgYDVR0PAQH/BAQDAgEGMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYE
# FI/wS3+oLkUkrk1Q+mOai97i3Ru8MA0GCSqGSIb3DQEBCwUAA4IBAQBLQNvAUKr+
# yAzv95ZURUm7lgAJQayzE4aGKAczymvmdLm6AC2upArT9fHxD4q/c2dKg8dEe3jg
# r25sbwMpjjM5RcOO5LlXbKr8EpbsU8Yt5CRsuZRj+9xTaGdWPoO4zzUhw8lo/s7a
# wlOqzJCK6fBdRoyV3XpYKBovHd7NADdBj+1EbddTKJd+82cEHhXXipa0095MJ6RM
# G3NzdvQXmcIfeg7jLQitChws/zyrVQ4PkX4268NXSb7hLi18YIvDQVETI53O9zJr
# lAGomecsMx86OyXShkDOOyyGeMlhLxS67ttVb9+E7gUJTb0o2HLO02JQZR7rkpeD
# MdmztcpHWD9fMIIEFDCCAvygAwIBAgILBAAAAAABL07hUtcwDQYJKoZIhvcNAQEF
# BQAwVzELMAkGA1UEBhMCQkUxGTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYtc2ExEDAO
# BgNVBAsTB1Jvb3QgQ0ExGzAZBgNVBAMTEkdsb2JhbFNpZ24gUm9vdCBDQTAeFw0x
# MTA0MTMxMDAwMDBaFw0yODAxMjgxMjAwMDBaMFIxCzAJBgNVBAYTAkJFMRkwFwYD
# VQQKExBHbG9iYWxTaWduIG52LXNhMSgwJgYDVQQDEx9HbG9iYWxTaWduIFRpbWVz
# dGFtcGluZyBDQSAtIEcyMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA
# lO9l+LVXn6BTDTQG6wkft0cYasvwW+T/J6U00feJGr+esc0SQW5m1IGghYtkWkYv
# maCNd7HivFzdItdqZ9C76Mp03otPDbBS5ZBb60cO8eefnAuQZT4XljBFcm05oRc2
# yrmgjBtPCBn2gTGtYRakYua0QJ7D/PuV9vu1LpWBmODvxevYAll4d/eq41JrUJEp
# xfz3zZNl0mBhIvIG+zLdFlH6Dv2KMPAXCae78wSuq5DnbN96qfTvxGInX2+ZbTh0
# qhGL2t/HFEzphbLswn1KJo/nVrqm4M+SU4B09APsaLJgvIQgAIMboe60dAXBKY5i
# 0Eex+vBTzBj5Ljv5cH60JQIDAQABo4HlMIHiMA4GA1UdDwEB/wQEAwIBBjASBgNV
# HRMBAf8ECDAGAQH/AgEAMB0GA1UdDgQWBBRG2D7/3OO+/4Pm9IWbsN1q1hSpwTBH
# BgNVHSAEQDA+MDwGBFUdIAAwNDAyBggrBgEFBQcCARYmaHR0cHM6Ly93d3cuZ2xv
# YmFsc2lnbi5jb20vcmVwb3NpdG9yeS8wMwYDVR0fBCwwKjAooCagJIYiaHR0cDov
# L2NybC5nbG9iYWxzaWduLm5ldC9yb290LmNybDAfBgNVHSMEGDAWgBRge2YaRQ2X
# yolQL30EzTSo//z9SzANBgkqhkiG9w0BAQUFAAOCAQEATl5WkB5GtNlJMfO7Fzko
# G8IW3f1B3AkFBJtvsqKa1pkuQJkAVbXqP6UgdtOGNNQXzFU6x4Lu76i6vNgGnxVQ
# 380We1I6AtcZGv2v8Hhc4EvFGN86JB7arLipWAQCBzDbsBJe/jG+8ARI9PBw+Dpe
# VoPPPfsNvPTF7ZedudTbpSeE4zibi6c1hkQgpDttpGoLoYP9KOva7yj2zIhd+wo7
# AKvgIeviLzVsD440RZfroveZMzV+y5qKu0VN5z+fwtmK+mWybsd+Zf/okuEsMaL3
# sCc2SI8mbzvuTXYfecPlf5Y1vC0OzAGwjn//UYCAp5LUs0RGZIyHTxZjBzFLY7Df
# 8zCCBJ8wggOHoAMCAQICEhEh1pmnZJc+8fhCfukZzFNBFDANBgkqhkiG9w0BAQUF
# ADBSMQswCQYDVQQGEwJCRTEZMBcGA1UEChMQR2xvYmFsU2lnbiBudi1zYTEoMCYG
# A1UEAxMfR2xvYmFsU2lnbiBUaW1lc3RhbXBpbmcgQ0EgLSBHMjAeFw0xNjA1MjQw
# MDAwMDBaFw0yNzA2MjQwMDAwMDBaMGAxCzAJBgNVBAYTAlNHMR8wHQYDVQQKExZH
# TU8gR2xvYmFsU2lnbiBQdGUgTHRkMTAwLgYDVQQDEydHbG9iYWxTaWduIFRTQSBm
# b3IgTVMgQXV0aGVudGljb2RlIC0gRzIwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCwF66i07YEMFYeWA+x7VWk1lTL2PZzOuxdXqsl/Tal+oTDYUDFRrVZ
# UjtCoi5fE2IQqVvmc9aSJbF9I+MGs4c6DkPw1wCJU6IRMVIobl1AcjzyCXenSZKX
# 1GyQoHan/bjcs53yB2AsT1iYAGvTFVTg+t3/gCxfGKaY/9Sr7KFFWbIub2Jd4NkZ
# rItXnKgmK9kXpRDSRwgacCwzi39ogCq1oV1r3Y0CAikDqnw3u7spTj1Tk7Om+o/S
# WJMVTLktq4CjoyX7r/cIZLB6RA9cENdfYTeqTmvT0lMlnYJz+iz5crCpGTkqUPqp
# 0Dw6yuhb7/VfUfT5CtmXNd5qheYjBEKvAgMBAAGjggFfMIIBWzAOBgNVHQ8BAf8E
# BAMCB4AwTAYDVR0gBEUwQzBBBgkrBgEEAaAyAR4wNDAyBggrBgEFBQcCARYmaHR0
# cHM6Ly93d3cuZ2xvYmFsc2lnbi5jb20vcmVwb3NpdG9yeS8wCQYDVR0TBAIwADAW
# BgNVHSUBAf8EDDAKBggrBgEFBQcDCDBCBgNVHR8EOzA5MDegNaAzhjFodHRwOi8v
# Y3JsLmdsb2JhbHNpZ24uY29tL2dzL2dzdGltZXN0YW1waW5nZzIuY3JsMFQGCCsG
# AQUFBwEBBEgwRjBEBggrBgEFBQcwAoY4aHR0cDovL3NlY3VyZS5nbG9iYWxzaWdu
# LmNvbS9jYWNlcnQvZ3N0aW1lc3RhbXBpbmdnMi5jcnQwHQYDVR0OBBYEFNSihEo4
# Whh/uk8wUL2d1XqH1gn3MB8GA1UdIwQYMBaAFEbYPv/c477/g+b0hZuw3WrWFKnB
# MA0GCSqGSIb3DQEBBQUAA4IBAQCPqRqRbQSmNyAOg5beI9Nrbh9u3WQ9aCEitfhH
# NmmO4aVFxySiIrcpCcxUWq7GvM1jjrM9UEjltMyuzZKNniiLE0oRqr2j79OyNvy0
# oXK/bZdjeYxEvHAvfvO83YJTqxr26/ocl7y2N5ykHDC8q7wtRzbfkiAD6HHGWPZ1
# BZo08AtZWoJENKqA5C+E9kddlsm2ysqdt6a65FDT1De4uiAO0NOSKlvEWbuhbds8
# zkSdwTgqreONvc0JdxoQvmcKAjZkiLmzGybu555gxEaovGEzbM9OuZy5avCfN/61
# PU+a003/3iCOTpem/Z8JvE3KGHbJsE2FUPKA0h0G9VgEB7EYMIIEpzCCA4+gAwIB
# AgIOSBtqB6lCTB6q/vPN8Q8wDQYJKoZIhvcNAQELBQAwTDEgMB4GA1UECxMXR2xv
# YmFsU2lnbiBSb290IENBIC0gUjMxEzARBgNVBAoTCkdsb2JhbFNpZ24xEzARBgNV
# BAMTCkdsb2JhbFNpZ24wHhcNMTYwNjE1MDAwMDAwWhcNMjQwNjE1MDAwMDAwWjBu
# MQswCQYDVQQGEwJCRTEZMBcGA1UEChMQR2xvYmFsU2lnbiBudi1zYTFEMEIGA1UE
# AxM7R2xvYmFsU2lnbiBFeHRlbmRlZCBWYWxpZGF0aW9uIENvZGVTaWduaW5nIENB
# IC0gU0hBMjU2IC0gRzMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDZ
# t7olrZTzW75BBhxTygwQjFFBWTN5ZPRXneTVJcTsUIRYmHJ5QOIveNSS6iYOnq6V
# fPvE/XFE3YxftyOLXrv0/EvLIz3DdgP10YxFvHF1HYvSiYm+41E9xsiKsjE1B265
# 9bpqDfQQn67VYkkoe+xXuqsyfLF90qJWBjbusO/Qaq7qqx/WDZ98lvutcJktXZXw
# gNB5RuxVOszTOPsEB6gHdYKC4NB+d7iP69Io/K5tFGhBf3ZD10i6YETht3Lo0PAg
# A3va2rQGdceyA974lMZoj157npudNuDO0mvGxmvpFCK1cX62j1of2+du9EIQkGjm
# K0UQT3O6LNfFMWpy3WNzAgMBAAGjggFjMIIBXzAOBgNVHQ8BAf8EBAMCAQYwHQYD
# VR0lBBYwFAYIKwYBBQUHAwMGCCsGAQUFBwMJMBIGA1UdEwEB/wQIMAYBAf8CAQAw
# HQYDVR0OBBYEFNwsWCwqbzUtn3mVqEhdxG0+U7+5MB8GA1UdIwQYMBaAFI/wS3+o
# LkUkrk1Q+mOai97i3Ru8MD4GCCsGAQUFBwEBBDIwMDAuBggrBgEFBQcwAYYiaHR0
# cDovL29jc3AyLmdsb2JhbHNpZ24uY29tL3Jvb3RyMzA2BgNVHR8ELzAtMCugKaAn
# hiVodHRwOi8vY3JsLmdsb2JhbHNpZ24uY29tL3Jvb3QtcjMuY3JsMGIGA1UdIARb
# MFkwCwYJKwYBBAGgMgECMAcGBWeBDAEDMEEGCSsGAQQBoDIBXzA0MDIGCCsGAQUF
# BwIBFiZodHRwczovL3d3dy5nbG9iYWxzaWduLmNvbS9yZXBvc2l0b3J5LzANBgkq
# hkiG9w0BAQsFAAOCAQEAdgnEzC/Z7x5LqfhX80A5IcpMPB2eKSsg1CtE0ojOGg0F
# z4OBu+tpvDGNKsTHRMxgYJQcz6HhAiQOrVu+LMInHme36CgfMlHjOfOY37ifLosq
# tHsKA7y9NgSPydCcT6MCJ5mw8EXpNN/kOqO3BjfYbyp5kNTUTlhx7FOpYZj3OWng
# EpxXWHKGJymlHeUy8yuZl1q/K7A8tAbqDmTst81lgCQXwtk39bEmEDVHe5oCulSi
# RZP/eb8ajMWftZ/feOdrUPFHlGlLJLjaBegMnU8G7EoxIH5PXYaELzWjzZzBhFcf
# H63A4qSx7ylrIZem1P7tAzew/PWNKrzchIPj3sPnXzCCBcAwggSooAMCAQICDAOW
# 0gvTBA+nfptoEzANBgkqhkiG9w0BAQsFADBuMQswCQYDVQQGEwJCRTEZMBcGA1UE
# ChMQR2xvYmFsU2lnbiBudi1zYTFEMEIGA1UEAxM7R2xvYmFsU2lnbiBFeHRlbmRl
# ZCBWYWxpZGF0aW9uIENvZGVTaWduaW5nIENBIC0gU0hBMjU2IC0gRzMwHhcNMTkw
# MjI1MDg0MzE1WhcNMjIwMzI1MDkwNzIwWjCCARMxHTAbBgNVBA8MFFByaXZhdGUg
# T3JnYW5pemF0aW9uMRMwEQYDVQQFEwpIUkIgMTAwNTIxMRMwEQYLKwYBBAGCNzwC
# AQMTAkRFMR4wHAYLKwYBBAGCNzwCAQITDU5pZWRlcnNhY2hzZW4xHTAbBgsrBgEE
# AYI3PAIBARMMQnJhdW5zY2h3ZWlnMQswCQYDVQQGEwJERTEWMBQGA1UECBMNTmll
# ZGVyc2FjaHNlbjESMBAGA1UEBxMJV29sZnNidXJnMRwwGgYDVQQJExNNYXgtdm9u
# LUxhdWUtV2VnIDE5MRgwFgYDVQQKEw9URS1TWVNURU1TIEdtYkgxGDAWBgNVBAMT
# D1RFLVNZU1RFTVMgR21iSDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
# AOlrV4R2g7Iz6/DBAaoovU8Ot9hkt7sQYCs6BQmc+9OmM0hWxFvFO4EMCx7ba8kU
# 6PM2huLkVwLyLYBaBb8YQEo3oW/L34RbFpQHIpN+TCtipCgJy4t9EaLooXkdB0xb
# Q8D8fIle+oZmnfc0f+fPQoNBE5zWW+WrY4cguDu/XeIu+T6S6w3V/YrYNlVsjTi4
# xaIz99N53hJekiudlm9qlXH65H1spVHkrUzhf1zn3g2Zn1gozs8zNiIsX+oRUSnz
# 5gzzrjs4g9U/6N9VD3R5q8U64AGhwm9/N1eX4Vx2s41Qlq4rGZ5aFu3+KAP4KnWT
# zigJ+wIkgOAG/nWga+dNF4ECAwEAAaOCAbUwggGxMA4GA1UdDwEB/wQEAwIHgDCB
# oAYIKwYBBQUHAQEEgZMwgZAwTgYIKwYBBQUHMAKGQmh0dHA6Ly9zZWN1cmUuZ2xv
# YmFsc2lnbi5jb20vY2FjZXJ0L2dzZXh0ZW5kY29kZXNpZ25zaGEyZzNvY3NwLmNy
# dDA+BggrBgEFBQcwAYYyaHR0cDovL29jc3AyLmdsb2JhbHNpZ24uY29tL2dzZXh0
# ZW5kY29kZXNpZ25zaGEyZzMwVQYDVR0gBE4wTDBBBgkrBgEEAaAyAQIwNDAyBggr
# BgEFBQcCARYmaHR0cHM6Ly93d3cuZ2xvYmFsc2lnbi5jb20vcmVwb3NpdG9yeS8w
# BwYFZ4EMAQMwCQYDVR0TBAIwADBFBgNVHR8EPjA8MDqgOKA2hjRodHRwOi8vY3Js
# Lmdsb2JhbHNpZ24uY29tL2dzZXh0ZW5kY29kZXNpZ25zaGEyZzMuY3JsMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSddGUb1NG8k/OYyFUq59Rr79uzHzAf
# BgNVHSMEGDAWgBTcLFgsKm81LZ95lahIXcRtPlO/uTANBgkqhkiG9w0BAQsFAAOC
# AQEACRJ6PggfKOL77yvNH8Fmk72/cBsQ9pHR1t6yOuILJonG54xRe5/Uff183a6M
# I189A9VDsWn9yQj76iTTDkL2SdRxzHmw9aXDaLrpVE5I7CnM8eIBvBmY2yJtzsGF
# sVbFmLbAwNxnbD9ooymxI7hqUi+CziGGNEvPMgqasl/haRz0Eal6ZJllZaRw/BQi
# /xoWq3OWzUfPhyg7rQpxCB9Xb91Rsej/IHq5GYTAxgGi1vUKIT+P/JeJCF4DJyfC
# 3zNxzmzGb9WHOmpIt4AfADnb7w/CDLrJ+OJ0uUceMueqGZVZTes/QgVaokFbwp3e
# vwvUQucemIS7x42Hox60rbNBuzGCBMUwggTBAgEBMH4wbjELMAkGA1UEBhMCQkUx
# GTAXBgNVBAoTEEdsb2JhbFNpZ24gbnYtc2ExRDBCBgNVBAMTO0dsb2JhbFNpZ24g
# RXh0ZW5kZWQgVmFsaWRhdGlvbiBDb2RlU2lnbmluZyBDQSAtIFNIQTI1NiAtIEcz
# AgwDltIL0wQPp36baBMwCQYFKw4DAhoFAKB4MBgGCisGAQQBgjcCAQwxCjAIoAKA
# AKECgAAwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEO
# MAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFPq7ak8TNsicwU7nspXEtSej
# +LD6MA0GCSqGSIb3DQEBAQUABIIBAJGzVHxSNvkTBO+RBP7KM5piSEb/4ZKyyMRv
# 5ffU0lEWOpW1Dsm8YxUqtTGpIaBtN16okQyT8u82Zl92g9dOllFCQ+dA2AyUepvn
# cIiN2sPwCOLKVINfVaKW80A7eZYRSNOUggOnF1qLcz2lo2olBxozv0MP/2xy4AW3
# 3Y9hhqGzSNLL9fWUn9pX5yGtijsgCPGd1z49srfx/55pAXcvUg4ncwXF16GwahVl
# Tc97taCQ5IylXf0b0bGeoIMTV6fwNwlNZARlVYuXkZVrRD91ThIHsbR3p6J6ausc
# 6/VfRAUMXBvFENtXXYI1rPZ0QkmLzjQDI8AWcwU1kpoxp1QdzMuhggKiMIICngYJ
# KoZIhvcNAQkGMYICjzCCAosCAQEwaDBSMQswCQYDVQQGEwJCRTEZMBcGA1UEChMQ
# R2xvYmFsU2lnbiBudi1zYTEoMCYGA1UEAxMfR2xvYmFsU2lnbiBUaW1lc3RhbXBp
# bmcgQ0EgLSBHMgISESHWmadklz7x+EJ+6RnMU0EUMAkGBSsOAwIaBQCggf0wGAYJ
# KoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0BCQUxDxcNMjAwODE4MTI1
# MTIzWjAjBgkqhkiG9w0BCQQxFgQUXoxBzlmZGOa1EdUAjqwqVQWr9KIwgZ0GCyqG
# SIb3DQEJEAIMMYGNMIGKMIGHMIGEBBRjuC+rYfWDkJaVBQsAJJxQKTPseTBsMFak
# VDBSMQswCQYDVQQGEwJCRTEZMBcGA1UEChMQR2xvYmFsU2lnbiBudi1zYTEoMCYG
# A1UEAxMfR2xvYmFsU2lnbiBUaW1lc3RhbXBpbmcgQ0EgLSBHMgISESHWmadklz7x
# +EJ+6RnMU0EUMA0GCSqGSIb3DQEBAQUABIIBAB9gD7tQIwEJfX3BfBnxyV7ymG1u
# UxPSGt+35D/cmbVFvQHjzTIADRqcSsd31zUPMyYduxOVZ9f//kWU6dLlg+5Rs6QC
# 6t2JbtjFT7umBBjrkBiIUOD7UfgbIHr8ZTDptvoMyS3fg+B8FxWi8oG+L9CZUy2N
# YycacMAxyGiPpQkaOmabmt6UYyUU1NxzZF6i0FqMuElZ8KrxBi7y0W9zhCj2eQRw
# eOTKBcl0sM+D/pfHYqZoXYgyNJquUO3Mwqa6dJASjXZE8xcbtVwOZjc2cRBw5Sxd
# YUIm4bjg4eGec7Y2sXn/hvl4gfxaq2/+s4hELKk+TlAtmh8qAv5n6fu5qH4=
# SIG # End signature block
